/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "loc_list.hh"
#include <iostream>
#include <stdexcept>
using namespace std;

using namespace sends;
using thread::semlock;


//======================================  Constructor
loc_list::loc_list(void) 
    : _scan_time(0)
{}

//======================================  Destructor
loc_list::~loc_list(void) {
}

//======================================  Archive level names
const char* 
loc_list::arch_name(arch_type a) {
    switch (a) {
    case kOnline: 
	return "online";
	break;
    case kCache: 
	return "cache";
	break;
    case kArchive: 
	return "archive";
	break;
    case kOffline: 
	return "offline";
    default:
	return "";
    }
}

//======================================  Coalesce segments
void
loc_list::coalesce(void) {
    if (empty()) return;
    semlock(_mutex, semlock::wrlock);
    iterator inx = begin();
    for (const_iterator jnx = ++begin(); jnx != end(); jnx++) {
	if (!inx->merge(*jnx)) {
	    inx++;
	    if (inx != jnx) *inx = *jnx;
	}
    }
    ++inx;
    if (inx != end()) this->erase(inx, end());
}

//======================================  Copy the last frame_loc in the list
bool
loc_list::find_back(frame_loc& fl) const {
    if (empty()) return false;
    semlock(_mutex, semlock::rdlock);
    fl = back();
    return true;
}

//======================================  Copy the first frame_loc in the list
bool
loc_list::find_front(frame_loc& fl) const {
    if (empty()) return false;
    semlock(_mutex, semlock::rdlock);
    fl = front();
    return true;
}

#ifdef LOC_LIST_LOCKED_FINDS
//======================================  Find frame_loc for a specified gps
const frame_loc* 
loc_list::find_loc(gps_type gps) const {
    if (empty()) return 0;
    read_lock();
    if (archive() == kOnline) return &back();
    unsigned long inx = upper_bound(gps);
    if (inx != size() && (*this)[inx].in_range(gps)) {
	return &((*this)[inx]);
    }
    unlock();
    return 0;
}

//======================================  Find frame_loc for a specified gps
frame_loc* 
loc_list::find_loc(gps_type gps) {
    if (empty()) return 0;
    write_lock();
    if (archive() == kOnline) return &back();
    unsigned long inx = upper_bound(gps);
    if (inx != size() && (*this)[inx].in_range(gps)) {
	return &((*this)[inx]);
    }
    unlock();
    return 0;
}
#endif

//======================================  Find frame_loc for a specified gps
bool
loc_list::find_loc(gps_type gps, frame_loc& fl) const {
    if (archive() == kOnline) {
	return find_back(fl);
    } else if (!empty()) {
	semlock(_mutex, semlock::rdlock);
	unsigned long inx = upper_bound(gps);
	if (inx != size() && (*this)[inx].in_range(gps)) {
	    fl = (*this)[inx];
	    return true;
	}
    }
    return false;
}

//======================================  Get the archive level type code
loc_list::arch_type
loc_list::get_arch_type(const string& arch) {
    if (arch == "online")  return kOnline;
    if (arch == "cache")   return kCache;
    if (arch == "archive") return kArchive;
    if (arch == "offline") return kOffline;
    throw runtime_error("Unknown archive level.");
}

//=====================================  Test if  data exists in range.
bool
loc_list::in_range(gps_type gps) const {
    if (empty()) return false;
    if (archive() == kOnline) return true;
    semlock(_mutex, semlock::rdlock);
    unsigned long inx = upper_bound(gps);
    if (inx != size() && (*this)[inx].in_range(gps)) return true;
    return false;
}

//======================================  Count memory space
index_type
loc_list::mem_use(void) const {
    index_type total = sizeof(loc_list) + _scan_path.size();
    index_type N = size();
    for (index_type i=0; i<N; i++) {
	total += at(i).mem_use();
    }
    total += (capacity() - size()) * sizeof(frame_loc);
    return total;
}

//======================================  Time range covered by this list
epoch
loc_list::range(void) const {
    semlock lock(_mutex, semlock::rdlock);
    epoch rv(0,0);
    for (size_t i=0; i<size(); i++) {
	rv.merge(epoch(at(i).start(), at(i).stop()));
    }
    return rv;
}
    
//======================================  Lock for read access
void
loc_list::read_lock(void) const {
    //cout << "read lock ";
    //if ( !empty()) cout << frame_loc::arch_name(begin()->archive()) 
    //			  << ":" <<  begin()->prefix();
    //cout << endl;
    _mutex.readlock();
}

//======================================  Set location archive type
void 
loc_list::set_archive(arch_type a) {
    _arch_type = a;
}

//======================================  Set the scan path.
void 
loc_list::set_scan_path(const std::string& path) {
    _scan_path = path;
}

//======================================  Set the (re)scan gps time.
void 
loc_list::set_scan_time(gps_type gps) {
    _scan_time = gps;
}

//======================================  Lock for read access
void
loc_list::swap_vec(loc_list& l) {
    if (this == &l) return;
    semlock lock1(l._mutex, semlock::wrlock);
    semlock lock2(  _mutex, semlock::wrlock);
    std::vector<frame_loc>::swap(l);
}

//======================================  Lock for read access
void
loc_list::unlock(void) const {
    //cout << "unlock ";
    //if ( !empty()) cout << frame_loc::arch_name(begin()->archive()) 
    //			  << ":" <<  begin()->prefix();
    //cout << endl;
    _mutex.unlock();
}

//======================================  Return index to uer bound
//
//  The upper bound is the lowest entry for which the stop time is greater 
//  than the specified gps.
//
index_type
loc_list::upper_bound(gps_type gps) const {
    index_type lo = 0;
    index_type hi = size();
    while (lo < hi) {
	index_type mid = (hi + lo)/2;
	if (gps < (*this)[mid].stop()) hi = mid;
	else if (mid == lo)      break;
	else                     lo = mid;
    }
    return hi;
}

//======================================  Lock for read access
void
loc_list::write_lock(void) {
    //cout << "write lock ";
    //if ( !empty()) cout << frame_loc::arch_name(begin()->archive()) 
    //			  << ":" << begin()->prefix();
    //cout << endl;
    _mutex.writelock();
}
