/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "seda_stage.hh"
#include <iostream>
#include <stdexcept>
#include <cstring>
#include <time.h>
#include <errno.h>

using namespace sends;
using namespace std;

//======================================  Message mutex
thread::mutex sends::seda_stage::msg_mutex;

//======================================  Construct a stage pool
seda_stage::seda_stage(int nThread, int nTrans)
    : thread_pool(nThread), dbglvl(dlvl_none), input(nTrans, stage_name()),
      output(0), alternate(0), error_queue(0), primary_count(0),
      alternate_count(0), reject_count(0)
{
}

//======================================  Destroy a stage pool
seda_stage::~seda_stage(void) {
    transaction* trans = 0;
    while (input.pop_if_present(trans)) {
	trans->release();
    }
}

//======================================  Stage name
const char*
seda_stage::stage_name(void) const {
    return "abstract";
}

//======================================  Time-stamped line
static ostream&
time_line(ostream& out) {
    struct tm now_tm;
    time_t now = time(0);
    localtime_r(&now, &now_tm);
    now_tm.tm_year += 1900;
    now_tm.tm_mon  += 1;
    out << now_tm.tm_year << "." << now_tm.tm_mon  << "." 
	<< now_tm.tm_mday << " " << now_tm.tm_hour << ":" 
	<< now_tm.tm_min  << ":" << now_tm.tm_sec  << " ";
    return out;
}

//======================================  Process transactions
void
seda_stage::thread_entry(void) {

    //----------------------------------  Initialize this thread.
    init_thread();

    //----------------------------------  Loop over transactions.
    while (test_run()) {
	transaction* trans = 0;
	try {
	    input.pop(trans);
	} catch (thread::mt_queue_shutdown& e) {
	    continue; // this exception will be handled by test_run
	}
	if (trans) {
	    debug_msg(dlvl_transaction, trans->ident(), "Start processing");
	    busy();
	    try {
		filter(trans);
	    } catch (std::overflow_error& e) {
		reject(trans, e.what());
	    }
	    done();
	} else {
	    debug_msg(dlvl_meta, -1, "Empty pop");
	}
    }
}

//======================================  Reject a transaction.
void 
seda_stage::cancel(transaction* trans, const std::string& msg) {
    reject_count++;
    error_msg(elvl_severe, *trans, msg);
    trans->release();
}

//======================================  Print a debug message
void 
seda_stage::debug_msg(dbg_level lvlmin, long tid, const string& msg) const {
    if (debug(lvlmin)) {
	thread::semlock lock(msg_mutex);
	time_line(cerr);
	cerr << "[" << stage_name() << ":" << tid << "] " << msg << endl;
    }
}

//======================================  Print an error message
void 
seda_stage::error_msg(err_level elvl, const transaction& t, 
		      const string& msg) const {
    thread::semlock lock(msg_mutex);
    time_line(cerr);
    cerr << "[" << stage_name() << ":" << t.ident() << "] ";
    switch (elvl) {
    case elvl_log:
	break;
    case elvl_warning:
	cerr << "warning: ";
	break;
    case elvl_error:
	cerr << "error: ";
	break;
    case elvl_severe:
	cerr << "severe: ";
	break;
    }
    cerr << msg << endl;
}

//======================================  Link output to the input of pool
void 
seda_stage::init_thread(void) {
}

//======================================  Link alternate to the input of pool
void 
seda_stage::link_alternate(seda_stage& pool) {
    alternate = &pool.ref_input();
}

//======================================  Link output to the input of pool
void 
seda_stage::link_error(seda_stage& pool) {
    error_queue = &pool.ref_input();
}

//======================================  Link output to the input of pool
void 
seda_stage::link_output(seda_stage& pool) {
    output = &pool.ref_input();
}

//======================================  Pass transaction to alternate stage
void 
seda_stage::put_alternate(transaction* trans) {
    if (!alternate) throw logic_error("Alternate queue not defined.");
    string msg = "Send to alternate queue (";
    msg += alternate->name() + ")";
    debug_msg(dlvl_transaction, trans->ident(), msg);
    if (!alternate->push_if_not_full(trans)) {
	msg = "Stage: ";
	msg += stage_name();
	msg += " alternate queue is full.";
	throw overflow_error(msg);
    }
    alternate_count++;
}

//======================================  Pass transaction to next stage
void 
seda_stage::put_error(transaction* trans) {
    string msg = "Send to error queue (";
    trans->set_stage(0);
    int transid = trans->ident();
    if (!error_queue || !error_queue->push_if_not_full(trans)) {
	msg += trans->home_queue();
	trans->release();
    } else {
	msg += error_queue->name();
    }
    msg += ")";
    debug_msg(dlvl_transaction, transid, msg);
}

//======================================  Pass transaction to next stage
void 
seda_stage::put_output(transaction* trans) {
    if (!output) {
	throw logic_error("Output queue not defined.");
    }
    string msg = "Send to output queue (";
    msg += output->name() + ")";
    debug_msg(dlvl_transaction, trans->ident(), msg);
    if (!output->push_if_not_full(trans)) {
	msg = "Stage: ";
	msg += stage_name();
	msg += " output queue is full.";
	throw overflow_error(msg);
    }
    primary_count++;
}

//======================================  Reject a transaction.
void 
seda_stage::reject(transaction* trans, const std::string& msg) {
    trans->reset();
    reject_count++;
    error_msg(elvl_error, *trans, msg);
    put_error(trans);
}

//======================================  Set the debug level
void 
seda_stage::set_debug(dbg_level lvl) {
    dbglvl = lvl;
}

//======================================  Set stage parameters.
void 
seda_stage::set_parameters(const stage_pars& p) {
    set_debug(p._debug);
    input.resize(p._inq_capacity);
    set_size(p._threads);
    if (debug(dlvl_detail)) {
	cout << "Stage: " << stage_name() << " parameters set: debug("
	     << dbg_lvl_name(dbglvl) << ") input queue length(" 
	     << input.capacity() << ") pool size(" << capacity() << ")." 
	     << endl;
    }
}

//======================================  Shut down all the threads
void 
seda_stage::shut(void) {
    time_line(cout) << stage_name() << " stage statistics:" << endl;
    cout << "  Maximum threads:  " << capacity() << endl;
    cout << "  Current threads:  " << size()     << endl;
    cout << "  Busy threads:     " << get_busy() << endl;
    cout << "  Max busy threads: " << get_max_busy() << endl;
    cout << "  Events Processed: " << count()    << endl;
    cout << "  Events to Output: " << primary_count << endl;
    cout << "  Alternate events: " << alternate_count << endl;
    cout << "  Events rejected:  " << reject_count << endl;
    cout << "  Input-Q capacity: " << input.capacity() << endl;
    cout << "  Input-Q contents: " << input.size() << endl;
    cout << "  Input-Q waiting:  " << input.wait_count() << endl;
    set_size(0);
    input.shut();
    wait_empty();
    transaction* trans(0);
    while (input.pop_if_present(trans)) {
	trans->release();
    }
}

//======================================  Start all the threads
void 
seda_stage::start(void) {
    make_threads();
}

//======================================  Print error message with system code
void 
seda_stage::systm_msg(err_level elvl, const transaction& t, 
		      const string& msg) const {
    string sysmsg = msg;
    sysmsg += ": ";
    sysmsg += strerror(errno);
    error_msg(elvl, t, sysmsg);
}

//======================================  Convert debug level to name
const char* 
seda_stage::dbg_lvl_name(dbg_level code) {
    switch (code) {
    case dlvl_none:
	return "none";
    case dlvl_transaction:
	return "transaction";
    case dlvl_meta:
	return "meta";
    case dlvl_detail:
	return "detail";
    case dlvl_data:
	return "data";
    }
    return "junk";
}

//======================================  Convert debug level name to code
#define TEST_NAME(x) if (name ==dbg_lvl_name(x)) return x
seda_stage::dbg_level
seda_stage::dbg_lvl_code(const string& name) {
    TEST_NAME(dlvl_none);
    TEST_NAME(dlvl_transaction);
    TEST_NAME(dlvl_meta);
    TEST_NAME(dlvl_detail);
    TEST_NAME(dlvl_data);
    return dlvl_none;
}
#undef TEST_NAME
